from typing import Union, Optional

from flask import Flask, Response, make_response, redirect, request, session
from flask_oauthlib.client import OAuthRemoteApp
from flask_restplus import Namespace
from gitlab import Gitlab
from gitlab.v4.objects import User, CurrentUser

from portal import gitlab_factory, logger, oauth
from portal.rest import rest_api
from portal.rest.custom_resource import CustomResource

log = logger.get_logger(__name__)

oauth_namespace = Namespace('oauth')


def gitlab_enabled(app: Flask) -> bool:
    return app.config.get('GITLAB_URL') and \
           app.config.get('GITLAB_CLIENT_ID') and \
           app.config.get('GITLAB_CLIENT_SECRET')


def register_gitlab_app() -> Union[OAuthRemoteApp, None]:
    if not gitlab_enabled(oauth.app):
        return

    base_url = oauth.app.config['GITLAB_URL']
    client_id = oauth.app.config['GITLAB_CLIENT_ID']
    client_secret = oauth.app.config['GITLAB_CLIENT_SECRET']
    api_base_url = base_url + '/api/v4/'

    params = dict(
        client_id=client_id,
        client_secret=client_secret,
        base_url=api_base_url,
        request_token_url=None,
        access_token_url=base_url + "/oauth/token",
        authorize_url=base_url + "/oauth/authorize",
        access_token_method='POST',
    )

    oauth.register('gitlab', **params)


register_gitlab_app()


@oauth_namespace.route('/login')
class OAuthLogin(CustomResource):
    def get(self):
        if not gitlab_enabled(oauth.app):
            return {'message': 'Gitlab OAuth is not enabled'}, 404

        callback = rest_api.url_for(OAuthLoginAuthorized, _external=True)

        log.debug(f"Callback set: {callback}")
        return oauth.gitlab.authorize_redirect(callback)


@oauth_namespace.route('/login/authorized')
class OAuthLoginAuthorized(CustomResource):
    def get(self):
        if not gitlab_enabled(oauth.app):
            return {'message': 'Gitlab OAuth is not enabled'}, 404
        resp = oauth.gitlab.authorize_access_token()
        log.debug(f"Access token get: {resp}")
        if resp is None:
            return 'Access denied: reason=%s error=%s' % (
                request.args['error'],
                request.args['error_description']
            )

        token = extract_token(resp)
        login = login_to_gitlab_with_token(token)
        log.debug(f"[GITLAB] Gitlab Login response: {login}")
        return login


def user_oauth_register(user_info):
    from portal.service.users import UserService
    users = UserService()
    new_user = users.create(
        uco=None,
        username=user_info.username,
        name=user_info.name,
        email=user_info.email,
        is_admin=False
    )
    log.info(f"[GITLAB] Created user={new_user}")
    return new_user


def user_login(user_info: CurrentUser) -> Response:
    from portal.service.find import FindService
    find = FindService()
    user = find.user(user_info.username, throws=False)
    if not user:
        user = user_oauth_register(user_info)
    resp = make_response(redirect(oauth.app.config.get('FRONTEND_URL')))
    resp.set_cookie('username', user.username)
    return resp


def extract_token(resp):
    token = resp['access_token']
    log.debug(f"[GITLAB] Received token: {token}")
    session['gitlab_token'] = (token, '')
    return token


def gitlab_get_user(token) -> Optional[User]:
    client: Gitlab = gitlab_factory.instance(oauth_token=token)
    try:
        client.auth()
        user = client.user
        return user
    except Exception as ex:
        log.error(f"[GITLAB] Error: {ex}")
        return None


def login_to_gitlab_with_token(token):
    log.debug(f"Logging in to gitlab with token: {token}")
    try:
        me = gitlab_get_user(token)
        login = user_login(me)
        login.set_cookie('gitlab_token', token)
        return login
    except Exception as ex:
        log.error(f"[GITLAB] Error: {ex}")


