"""
Submissions service
"""
import datetime
import json
import logging
from pathlib import Path

from celery.result import AsyncResult
from werkzeug.utils import secure_filename

from portal import storage
from portal.async_celery import submission_processor, tasks
from portal.database import queries
from portal.database.models import Project, Submission, User
from portal.service import errors
from portal.service.general import GeneralService

log = logging.getLogger(__name__)


def upload_files_to_storage(file):
    filename = secure_filename(file.filename)
    uploads = storage.results.workspace_path / 'uploads'
    if not uploads.exists():
        uploads.mkdir(parents=True)
    path = uploads / filename
    file.save(str(path))
    log.info(f"[UPLOAD] Uploading file to {filename} to {path}")
    return path


def upload_file_is_allowed(file, throws: bool = True):
    extension = Path(file.filename).suffix
    log.debug(f"[ZIP] Extension for {file.filename}: {extension}")
    if not extension == '.zip':
        if throws:
            raise errors.PortalAPIError(
                400,
                f"File with extension \"{extension}\" is not allowed: #{file.filename}"
            )
        return False
    return True


def get_upload_file_path(file):
    upload_file_is_allowed(file)
    path = upload_files_to_storage(file)
    return path


class SubmissionsService(GeneralService):
    @property
    def _entity_klass(self):
        return Submission

    def _set_data(self, entity, **data):
        allowed = []
        return self.update_entity(entity, data, allowed=allowed)

    @property
    def storage_results(self):
        return self.storage.results.get(self.submission.id)

    def create(self, user: User = None, project: Project = None, submission_params: dict = None):
        """Creates a new submission in the database and downloads its files.

          Zip and git sources are processed differently.

          Args:
              user(User): the submission's author
              project(Project): the project the submission falls under
              submission_params(dict): custom parameters specified by
                  project with their values as a JSON-encoded string

          Returns(Submission): the created submission entity

          """
        # adding review at submission create is not supported
        submission_params = self.process_submission_params(submission_params, project, user=user)
        log.info(f"[SUBMIT] Create new submission for user {user.log_name} "
                 f"for project {project.log_name} "
                 f"with params: {submission_params}")
        new_submission = Submission(
            user=user, project=project, parameters=submission_params)
        self.write_entity(new_submission)

        # download relevant files into Storage
        self._entity = new_submission
        self.process_new_submission()
        return new_submission

    @property
    def submission(self) -> Submission:
        return self._entity

    @property
    def project(self) -> Project:
        return self.submission.project

    @property
    def storage(self):
        return storage

    def process_new_submission(self) -> AsyncResult:
        project = self.submission.project
        self.submission.parameters['file_params'] = project.config.file_whitelist
        self.write_entity(self.submission)
        task = self.create_process_submission_task()
        return task

    def create_process_submission_task(self) -> AsyncResult:
        log.info(f"[SUBMIT] Processing task for {self.submission.log_name}")
        result: AsyncResult = tasks.process_submission.delay(self.submission.id)
        self.submission.storage_task_id = result.task_id
        return result

    def copy_submission(self, note: str) -> Submission:
        """Copies a submission. Used at resubmitting

        Args:
            note(str): Note for the submission. Typically contains a reason for the resubmit.

        Returns(Submission): Copied submission
        """
        log.info(f"[COPY] Submission {self.submission.log_name} to "
                 f"same project {self.submission.project.log_name}: {note}")
        new_submission = Submission(user=self.submission.user, project=self.project,
                                    parameters=self.submission.parameters)
        new_submission.note = note
        self.write_entity(new_submission)
        tasks.clone_submission_files.delay(
            source_id=self.submission.id, target_id=new_submission.id)
        return new_submission

    def resend_submission(self):
        """ Resends the submission to the worker
        """
        log.info(f"[SUBMIT] Resending submission task for {self.submission.log_name}")
        tasks.start_processing_submission.delay(self.submission.id)

    def delete(self):
        """Deletes a submission

        """
        # TODO: dispatch action to clean the storage
        super(SubmissionsService, self).delete()
        log.info(f"[DEL] Submission delete submission files: {self.submission.log_name}")
        st_entity = storage.submissions.get(self.submission.id)
        if st_entity:
            st_entity.delete()
        log.info(f"[DEL] Submission delete result files: {self.submission.log_name}")
        st_entity = storage.results.get(self.submission.id)
        if st_entity:
            st_entity.delete()

    def update_submission_state(self, **data) -> Submission:
        """Updates submission's state

        Args:
            data(dict): should contain one key, 'state' with the new value

        Returns(Submission): Updated submission
        """
        new_state = data['state']
        if self.submission.change_state(new_state):
            self.write_entity(self.submission)
            log.info(f"[UPDATE] Submission state {self.submission.log_name} "
                     f"to {self.submission.state}")
        else:
            log.warning("[UPDATE] Submission state change is not possible")
        return self.submission

    def cancel_submission(self):
        """Cancels the submission

        """
        processor = submission_processor.SubmissionProcessor(self.submission)
        processor.revoke_task()

    def upload_results_to_storage(self, file):
        path = get_upload_file_path(file)
        task = tasks.upload_results_to_storage.delay(self.submission.id, path=str(path))
        return task

    def find_all(self, *args, **kwargs):
        # if user:
        #     query = query.filter(Submission.user == user)
        #
        # if state:
        #     query = query.filter(Submission.state == state)
        #
        # if course:
        #     query = query.filter(Submission.course == course)
        #
        #     if projects:
        #         query = query.filter(Submission.project_id.in_(projects))
        #     if role_ids:
        #         query.join(Submission.course).join(Course.roles).filter(Role.id.in_(role_ids))
        #     if group_ids:
        #         query.join(Submission.course).join(Course.groups).filter(Group.id.in_(group_ids))
        # return filter_user_avail_submissions(query)
        return queries.list_submissions_for_user(**kwargs)

    def process_submission_params(self, params: dict, project: Project, user: User):
        file_params = params.get('file_params')
        source = file_params.get('source')
        if source and source['type'] == 'git':
            url = source.get('url')
            if not url:
                source['url'] = self.__get_default_git_url(project, user)
                file_params['from_dir'] = params.get('from_dir') or project.codename
        return params

    def __get_default_git_url(self, project: Project, user: User):
        git_base = self._config.get('GIT_REPO_BASE', 'git@gitlab.fi.muni.cz')
        return f"{git_base}/{user.username}/{project.course.codename}"

    def get_stats(self):
        """Gets the statistics for the submission
        Returns: JSON representation
        """
        results = self.storage_results
        if not results.path.exists():
            raise errors.ResultsAreNotUploaded(self.submission)
        stats = results.get('suite-stats.json')
        if not stats.exists():
            raise errors.SuiteStatsNotExists(self.submission)
        return json.loads(stats.read_text('utf-8'))

    def delete_cancelled_submissions(self):
        one_day = datetime.timedelta(hours=24)
        query_filter = queries.cancelled_submissions_for_deletion(one_day)
        log.info(f"[DEL] Cancelled submissions for deletion "
                 f"({query_filter}): {query_filter.all()}")

        for submission in query_filter.all():
            SubmissionsService(submission).delete()

    def archive_submissions(self):
        query_filter = queries.submissions_to_be_archived()
        log.info(f"[ARCH] Archiving submissions "
                 f"({query_filter}): {query_filter.all()}")
        for submission in query_filter.all():
            SubmissionsService(submission).archive()

    def archive(self):
        log.info(f"[ARCHIVE] Submission archive submission files: {self.submission.log_name}")
        st_entity = storage.submissions.get(self.submission.id)
        if st_entity:
            st_entity.clean()
        log.info(f"[ARCHIVE] Submission archive result files: {self.submission.log_name}")
        st_entity = storage.results.get(self.submission.id)
        if st_entity:
            st_entity.clean()


def nonempty_intersection(provided: list, required: list):
    if not required:
        return True
    return list(set(provided) & set(required))
